/*
 * sdmmc.c
*/

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
*/
/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/
/* --------------------------------------------------------------------------
  FILE      : sdmmc.c
  PROJECT   : TI Boot and Flash Utilities
  AUTHOR    : Daniel Allred
  DESC      : Generic SDMMC memory driver
-------------------------------------------------------------------------- */

// General type include
#include "tistdtypes.h"

// Device specific CSL
#include "device.h"

// Util functions
#include "util.h"

// The generic SD/MMC module driver 
#include "sdmmc.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/
static Uint8 Local_SDMMCInitCard(SDMMC_InfoHandle hSDMMCInfo);

/************************************************************
* Local Variable Definitions                                *
************************************************************/


/************************************************************
* Global Variable Definitions                               *
************************************************************/

#ifdef USE_IN_ROM
SDMMC_InfoObj gSDMMCInfo;
#endif
// Default SPI Config strcuture
const SDMMC_ConfigObj DEFAULT_SDMMC_CONFIG = 
{
SDMMC_LITTLE_ENDIAN,
SDMMC_LITTLE_ENDIAN,
SDMMC_DAT3_EDGE_DETECT_DISABLE,
(Uint8)FALSE,
(Uint8)FALSE,
(Uint8)FALSE,
(Uint8)0,
SDMMC_MODE_NATIVE, /* No SPI */
(Uint8)FALSE, /* Valid Only in SPI mode */
(Uint8)FALSE, /* Valid Only in SPI mode */
SDMMC_4BIT_DATABUS,
(Uint16)0xFF,
0xFFFF,
SDMMC_FIFO_LEVEL_32BYTES
};

SDMMC_ConfigHandle const hDEFAULT_SDMMC_CONFIG = (SDMMC_ConfigHandle) &DEFAULT_SDMMC_CONFIG;
#define SDMMC_BLOCK_LENGTH     (512)

/************************************************************
* Global Function Definitions                               *
************************************************************/

// Initialze SDMMC interface and find the details of the SD/MMC card used
SDMMC_InfoHandle SDMMC_open(Uint32 sdmmcPeripheralNum, SDMMC_ConfigHandle hSDMMCCfg)
{
  SDMMC_InfoHandle hSDMMCInfo;
  Uint32 localDataBytesPerBlk = 0;

  // Set SDMMCInfo handle
#ifdef USE_IN_ROM
  hSDMMCInfo = (SDMMC_InfoHandle) &gSDMMCInfo;
#else
  hSDMMCInfo = (SDMMC_InfoHandle) UTIL_allocMem(sizeof(SDMMC_InfoObj));
#endif

  if (hSDMMCCfg == NULL)
  {
    hSDMMCInfo->hSDMMCCfg = hDEFAULT_SDMMC_CONFIG;
  }
  else
  {
    hSDMMCInfo->hSDMMCCfg = hSDMMCCfg;
  }
  hSDMMCInfo->dataBytesPerBlk = SDMMC_BLOCK_LENGTH;

  localDataBytesPerBlk = hSDMMCInfo->dataBytesPerBlk;
  hSDMMCInfo->dataBytesPerBlkPower2 = 0;
  while(localDataBytesPerBlk > 1)
  {
    localDataBytesPerBlk >>= 1;
	hSDMMCInfo->dataBytesPerBlkPower2++;
  }
  // Open SDMMC peripheral
  Local_SDMMCInitCard(hSDMMCInfo);  

  return hSDMMCInfo;
}




// Defining this macro for the build will cause write (flash) ability to be removed
// This can be used for using this driver as read-only for ROM code
#ifndef USE_IN_ROM    

// Global Erase NOR Flash
Uint32 SDMMC_globalErase(SDMMC_InfoHandle hSDMMCInfo)
{
  return E_PASS;
}

// NAND Flash erase block function
Uint32 SDMMC_eraseBytes(SDMMC_InfoHandle hSDMMCInfo, Uint32 startAddr, Uint32 byteCnt)
{  
  return E_PASS;
}
#endif

/************************************************************
* Local Function Definitions                                *
************************************************************/

void SDMMCClearResponse() {
  SDMMC->MMCRSP01 = 0x0;
  SDMMC->MMCRSP23 = 0x0;
  SDMMC->MMCRSP45 = 0x0;
  SDMMC->MMCRSP67 = 0x0;
  SDMMC->MMCCIDX &= 0xFFC0;

}

/**
    Send Command to MMC/ SD

    command  Command to be sent to MMC/ SD
    argument  Arguments to be sent with the command
    checkStatus  0: no status check, 1: status check
	statusBits  status bits to be checked in MMCST0 register, if 'checkStatus' = 1

    E_PASS, on completion
*/
Uint8 SDMMCSendCmd(Uint32 command, Uint32 argument, Bool checkStatus) {
  
	Uint32 statusRegBits=0x0;
	volatile Uint32 dummyread;

	// Dummy read
	dummyread = SDMMC->MMCST0;

   	/* Clear Response Registers*/
   	SDMMCClearResponse();


   	//Setup the Argument Register and send CMD */
   	SDMMC->MMCARGHL = argument;
   	SDMMC->MMCCMD   = command;
  
   	/*Delay loop allowing cards to respond */
	UTIL_waitLoop(1000);
        
   	if (checkStatus == 1) {
	   	/* Wait for RspDne; exit on RspTimeOut or RspCRCErr
		   bit 2 - RSPDNE
		   bit 3 - 7 Error/timeout */
	   	while((statusRegBits & (0x00FC)) == 0 )
	    {
	     statusRegBits = SDMMC->MMCST0;
	    }

	    return( ((statusRegBits & 0x4) == 0x4) ? E_PASS : E_DEVICE );
	}

	return E_PASS;
}


/**
    Get Response from MMC/SD controller and Clear Response

    SDMMCResponse Command Response is received in this parameter

    if success, \c E_PASS, else error code

*/
Uint8 SDMMCGetResponse(SDMMC_ResponseData* sdMmcResponse) {

  sdMmcResponse->response[0] = (SDMMC->MMCRSP01 & 0xFFFF); 
  sdMmcResponse->response[1] = (SDMMC->MMCRSP01 & 0xFFFF0000) >> 16;
  sdMmcResponse->response[2] = (SDMMC->MMCRSP23 & 0xFFFF); 
  sdMmcResponse->response[3] = (SDMMC->MMCRSP23 & 0xFFFF0000) >> 16;
  sdMmcResponse->response[4] = (SDMMC->MMCRSP45 & 0xFFFF); 
  sdMmcResponse->response[5] = (SDMMC->MMCRSP45 & 0xFFFF0000) >> 16;
  sdMmcResponse->response[6] = (SDMMC->MMCRSP67 & 0xFFFF); 
  sdMmcResponse->response[7] = (SDMMC->MMCRSP67 & 0xFFFF0000) >> 16;
  sdMmcResponse->commandIdk  = (SDMMC->MMCCIDX  & 0x003F);
	
  SDMMCClearResponse();

  return E_PASS;
}

/**
	MMC/SD interface selection, Controller and Card Initialization

*/
static Uint8 Local_SDMMCInitCard(SDMMC_InfoHandle hSDMMCInfo) {
	Uint8 status;
	Uint32 i;
	Bool mmc = TRUE;
	Uint16 response;
	Uint32 temp, opTimeOut = SDMMC_OP_TIMEOUT;
	Uint16 cardReg[8];
	SDMMC_cardStatusReg cardStatus;

	/* Enable the Clock and set the divide-down value based on Silicon Operating Frequency
    During card identification process card operating frequency cannot exceed 400Khz*/

	/* Initialize to 0 assuming the sdHighCapacity card is not used */
	hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard = 0;
	

    /* Enable the clock and divide the VBUS_clk to get the clock going to the card to 
      150KHz Card clk = VBUS_CLK(24Mhz/4 (fixed divider)/(2*(19+1))
	  The clock as to be lower than 400Khz before reading CSD. Setting it as 300Khz to be
	  safe...13 = 0x13
    */
    SDMMC->MMCCLK =0x01FF;
    
    /*Delay required to card clock to stabilize */
    UTIL_waitLoop(10000); 
  
    /*Set the FIFO level(Threshold) 32 bytes (256 bit) always */
	SDMMC->MMCFIFOCTL |= ((Uint16)hSDMMCInfo->hSDMMCCfg->fifoThreshold) << 2; //0x0004;
   
    /* Put the CMD and Data Logic in RESET state 
   	   Read and write are little endian, DAT3 is disabled (default)
	   Buswidth = 4 bit */
    SDMMC->MMCCTL = 0x0007;
    SDMMC->MMCCTL = ((Uint16)hSDMMCInfo->hSDMMCCfg->busWidth) << 2; //0x0004;
    
	/* Max timeout set*/
    SDMMC->MMCTOR = (Uint16)hSDMMCInfo->hSDMMCCfg->timeoutResponse;// 0xFF;
    SDMMC->MMCTOD = (Uint16)hSDMMCInfo->hSDMMCCfg->timeoutRead; // 0xFFFF

   /* Card Identification cycle of MMC/SD */
   hSDMMCInfo->relCardAddress = 2; // Relative Card Address

   /* MMC needs some delay between two commands */
   UTIL_waitLoop(1000);

   /* Place all cards in idle state */
   status = SDMMCSendCmd( (0x4000 | SDMMC_GO_IDLE_STATE), SDMMC_STUFF_BITS, FALSE);
   if( status != E_PASS )
		return ( status );
    
   /* MMC needs some delay between two commands */
   UTIL_waitLoop(10000);  				
   
#ifdef SD_HIGH_CAPACITY_CARD

   /* CMD8 is accepted by both MMC and SD cards */
   /* Send CMD8 with RCA = 0 */
   status = SDMMCSendCmd( SDMMC_HICAP_CMD,(0x100|SD_HIGH_CAPACITY_ECHO_PATTERN),TRUE);

   /* MMC needs some delay between two commands */
   UTIL_waitLoop(1000);  				

   temp = (SDMMC->MMCRSP67 & 0x000000FF);

   if(temp !=  SD_HIGH_CAPACITY_ECHO_PATTERN) {
       status = E_DEVICE;
   }

#endif

#ifdef SD_HIGH_CAPACITY_CARD
   if( status != E_PASS ) {
#endif

   /* card is not SD high capacity card because it has not responded to SDHC card*/

   /* Place all cards in idle state */
   status = SDMMCSendCmd( (0x4000 | SDMMC_GO_IDLE_STATE), SDMMC_STUFF_BITS, FALSE);
   if( status != E_PASS )
		return ( status );
    
   /* MMC needs some delay between two commands */
   UTIL_waitLoop(1000);  				
   
	/* Send the operating Voltage Range */
	opTimeOut = SDMMC_OP_TIMEOUT;

   //Introduce a delay for slow cards
   UTIL_waitLoop(1000);  				
   
   /* CMD55 is accepted by both MMC and SD cards */
   /* Send CMD55 with RCA = 0 */
   status = SDMMCSendCmd( SDMMC_APP_CMD,0,TRUE);
   //Introduce a delay for slow cards
   UTIL_waitLoop(1000);  				

   // Assuming SD card. If this fails then MMC card will be tried inside IF
   /* Format and send cmd: Volt. window is (2.7-3.6v)*/
   status = SDMMCSendCmd( SD_APP_OP_COND, (0x00000000 | SDMMC_VDD_27_36), TRUE);
   
   if( status != E_PASS ){
   		// MMC
   		// Send command to check if the Operating Voltage Conditions of the MMC, 
		// match that of the host
		do{ 
	    	/* Format and send cmd: Volt. window is usually (3.4-3.2v)
	    	   Safe window is between 2.6V to 3.4V*/
	    	status = SDMMCSendCmd( MMC_SEND_OP_COND, (0x40000000 | SDMMC_VDD_27_36), TRUE);
			temp = (SDMMC->MMCRSP67 & 0xC0000000);
        	opTimeOut--;
		
		} while ( ((opTimeOut > 0)  && ((temp & 0x80000000) == 0)) );
		if(temp & 0x40000000) {
		   hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard = 1;
	    }
	    else {
		   hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard = 0;
	    }

   } else {
		// SD	
		temp = (SDMMC->MMCRSP67 & 0x80000000);
		if(temp != 0x80000000) {
		do{ 
            /* Send CMD55 with RCA = 0 for the next iteration */
            status = SDMMCSendCmd( SDMMC_APP_CMD,0,TRUE);

            //Introduce a delay for slow cards
            UTIL_waitLoop(1000);  				

            /* Format and send cmd: Volt. window is (2.7-3.6v)*/
            status = SDMMCSendCmd( SD_APP_OP_COND, (0x00000000 | SDMMC_VDD_27_36), TRUE);

            temp = (SDMMC->MMCRSP67 & 0x80000000);
							        
            opTimeOut--;

			} while ( ((opTimeOut > 0) && (temp == 0)) );
		}
						        
		mmc = FALSE;
     }
   
#ifdef SD_HIGH_CAPACITY_CARD
   }
   else {

	/* card is SD high capacity card because it has responded to SDHC card*/

	/* Send the operating Voltage Range */
	opTimeOut = SDMMC_OP_TIMEOUT;
   //Introduce a delay for slow cards
   UTIL_waitLoop(1000);  				
   
   /* CMD55 is accepted by both MMC and SD cards */
   /* Send CMD55 with RCA = 0 */
	status = SDMMCSendCmd( SDMMC_APP_CMD,0,TRUE);
   //Introduce a delay for slow cards
   UTIL_waitLoop(1000);  				

   // Assuming SD card. If this fails then MMC card will be tried inside IF
   /* Format and send cmd: Volt. window is (2.7-3.6v)*/
   status = SDMMCSendCmd( SD_APP_OP_COND, (0x40000000|(SDMMC_VDD_27_36)), TRUE);
    
   if( status != E_PASS ){
   		// MMC
   		// Send command to check if the Operating Voltage Conditions of the MMC, 
		// match that of the host
		do{ 
	    	/* Format and send cmd: Volt. window is usually (3.4-3.2v)
	    	   Safe window is between 2.6V to 3.4V*/
	    	status = SDMMCSendCmd( MMC_SEND_OP_COND, (0x40000000|(SDMMC_VDD_27_36)), TRUE);
			temp = (SDMMC->MMCRSP67 & 0xC0000000);
        	opTimeOut--;

		} while ( ((opTimeOut > 0) && ((temp & 0x80000000) == 0)) );
   
   } else {
		// SD	
		temp = (SDMMC->MMCRSP67 & 0xC0000000);
		if((temp & 0x80000000)!=0x80000000) {
			do{ 
			/* Send CMD55 with RCA = 0 for the next iteration */
				status = SDMMCSendCmd( SDMMC_APP_CMD,0,TRUE);
    
			    //Introduce a delay for slow cards
			    UTIL_waitLoop(1000);  				

				/* Format and send cmd: Volt. window is (2.7-3.6v)*/
				status = SDMMCSendCmd( SD_APP_OP_COND, (0x40000000|(SDMMC_VDD_27_36)), FALSE);

				temp = (SDMMC->MMCRSP67 & 0xC0000000);
							        
	    
        	opTimeOut--;

			} while ( ((opTimeOut > 0) && ((temp & 0x80000000) == 0)) );
		}

		mmc = FALSE;
   }
   
     if(temp & 0x40000000) {
	   hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard = 1;
	 }
	 else {
	   hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard = 0;
	 }
   }
#endif
   /* Return to the application, of the card is an MMC card or an SD card */
   cardStatus.multiMediaCard = (Bool)mmc;
   if(opTimeOut == 0)
     return E_FAIL;
      
   /*Delay between commands required for MMC */
   if(mmc == TRUE)
      UTIL_waitLoop(5000);    
   /* Ask all cards to send their CIDs */
   status = SDMMCSendCmd( SDMMC_ALL_SEND_CID, SDMMC_STUFF_BITS, TRUE);
   if( status != E_PASS )
   		return ( status );

   SDMMCGetResponse(&(hSDMMCInfo->sdMmcResponse));
   
   if( mmc == TRUE ){
		/* Set Relative Address for the card. As we are using only one card
		currently setting the card rca to  MMC_CARD_ID = 2
		We cannot use 1 because that is the default for all the SanDisk cards
		*/
		status = SDMMCSendCmd( MMC_SET_RELATIVE_ADDR, hSDMMCInfo->relCardAddress<<16, TRUE);

	  	if( status == E_PASS ) {
        	/* Response expected: R1 */
			response = hSDMMCInfo->sdMmcResponse.response[6];
	    	cardStatus.appSpecific = ( response&0xff );
	    	cardStatus.ready =  ( (response>> 8) & 1 );
	    	cardStatus.currentState = (SDMMC_CardStates) ( (response >> 9) & 0xf );
	    	cardStatus.eraseReset    = ( (response >> 13) & 1 );
	    	cardStatus.eccDisabled = ( (response >> 14) & 1 );
			cardStatus.wpEraseSkip = ( (response >> 15) & 1 );
	    	cardStatus.errorFlags   = hSDMMCInfo->sdMmcResponse.response[7];
      	} else {
			return E_FAIL;
		}
   } else {
		/* Read Relative Address from the card and store it in variable rca
		*/
		status = SDMMCSendCmd( SD_SEND_RELATIVE_ADDR, SDMMC_STUFF_BITS, TRUE);

	  	if( status == E_PASS ) {
        	/* Response expected: R1 */
			// RCA received on Interrupt Request - of winning card or of the host (rca =0)
        	hSDMMCInfo->relCardAddress = (SDMMC->MMCRSP67 & 0xFFFF0000) >> 16;
      	} else {
			return ( E_FAIL );
		}
   }
   
   /* Send Status may be required at this point */
   
   /* Read the CSD structure for MMC/SD */
   status = SDMMCSendCmd( SDMMC_SEND_CSD, hSDMMCInfo->relCardAddress<<16, TRUE);

   if( status == E_PASS ) {
    	/* Check whether the mode of operation is native MMC/SD or SPI */
   		SDMMCGetResponse(&(hSDMMCInfo->sdMmcResponse));

	    /* Read CSD from response R2 */
	    for( i=0; i<8; i++ ){
	  	  cardReg[i] = hSDMMCInfo->sdMmcResponse.response[i];
		}
	  	
    	// Fill the local structure with information required for other commands used in the driver
		hSDMMCInfo->mmcSdCSDRegInfo.permWriteProtect =  (cardReg[0]>>13) & 0x1;
		hSDMMCInfo->mmcSdCSDRegInfo.tmpWriteProtect =  (cardReg[0]>>12) & 0x1;
		hSDMMCInfo->mmcSdCSDRegInfo.writeBlkPartial =  (cardReg[1]>>5) & 0x1;
		hSDMMCInfo->mmcSdCSDRegInfo.writeBlkLenBytes =  1<<((cardReg[1]>>6) & 0xF);
    	hSDMMCInfo->mmcSdCSDRegInfo.wpGrpEnable = (cardReg[1]>>15) & 0x1;
    	hSDMMCInfo->mmcSdCSDRegInfo.wpGrpSize = (cardReg[2] & 0x7F) + 1;           ///> Extracting 7 bits: For MMC - 5 bits reqd; For SD - 7 bits reqd. (have to be taken care by user)
		hSDMMCInfo->mmcSdCSDRegInfo.dsrImp = (cardReg[4]>>12) &0x1;
		hSDMMCInfo->mmcSdCSDRegInfo.readBlkMisalign = (cardReg[4]>>13) &0x1;
		hSDMMCInfo->mmcSdCSDRegInfo.writeBlkMisalign = (cardReg[4]>>14) &0x1;
		hSDMMCInfo->mmcSdCSDRegInfo.readBlkPartial =  (cardReg[4]>>15) & 0x1;
		hSDMMCInfo->mmcSdCSDRegInfo.readBlkLenBytes =  1<<(cardReg[5] & 0xF);
    	hSDMMCInfo->mmcSdCSDRegInfo.sysSpecVersion = (cardReg[7]>>10) & 0xF;    ///> These bits are reserved in the case of SD card
   } else {
   		return E_FAIL;
   } 
   
   /* Select the Card which responded */
   /*  the rca value is sent as the upper 16 bits of the command argument  */
   status = SDMMCSendCmd( SDMMC_SELECT_CARD, hSDMMCInfo->relCardAddress<<16, TRUE);
   
   if( status == E_PASS ) {
		/* no timeout */
        if ( (hSDMMCInfo->relCardAddress) > 0 ) {
        	/*  the rca value is sent as the upper 16 bits of the command argument  */
	  	 	status = SDMMCSendCmd( SDMMC_SEND_STATUS, (hSDMMCInfo->relCardAddress)<<16, TRUE);
   		    if( status == E_PASS ) {
				SDMMCGetResponse(&(hSDMMCInfo->sdMmcResponse));
		    	/* Response expected: R1 in native mode*/
			    cardStatus.currentState = (SDMMC_CardStates) ( (hSDMMCInfo->sdMmcResponse.response[6] >> 9) & 0xf );
                if( (cardStatus.currentState != SDMMC_CARD_STATE_TRAN) && (cardStatus.currentState != SDMMC_CARD_STATE_PRG ) ) 
                	return ( E_FAIL );
           	}
        }
   } else {
   		return E_FAIL;
   }	

   // Setting MMC/SD always with bus_width = 4
   // Reset the timeout value
   opTimeOut = SDMMC_OP_TIMEOUT;
   
   if(mmc == FALSE) {
		// Setting bus width as 4 only for SD
		// For MMC and MMCPro it is always 1	
		do{
			/* Send CMD55 */
			status = SDMMCSendCmd( SDMMC_APP_CMD, (hSDMMCInfo->relCardAddress)<<16, TRUE);
		    if ( status != E_PASS )
				return ( E_FAIL );

			status = SDMMCSendCmd( SD_SET_BUS_WIDTH, 2, TRUE);

		} while( (status != E_PASS) && opTimeOut-- );
		 
		if( status != E_PASS )
		   return E_FAIL;
		
		// bit no.2 to 1 for 4 bit bus width
	  	SDMMC->MMCCTL |= 0x4;
   } else {
   		SDMMC->MMCCTL &= 0xFFFFFFFB;
   }
	
   /*Change the card clock close to max possible operating frequency */
	 
   /* Same for Silicon and QT */
   /* Silicon : Enable the clock and divide the VBUS_clk to get the clock going 
       	        to the card below 400KHz
		        Card clk = VBUS_CLK(24Mhz)/(2*(0+1))) ~ 12MhZ   
		QT    : Card clk = VBUS_CLK/(2)   
    */
   SDMMC->MMCCLK &=  0xFFFFFF00;

   return ( status );
}


/**
    Read N words from the MMC Controller Register

    data     -  Buffer for storing read data from the MMC
    numWords -  Number of words to be read from the MMC

*/
Uint8 SDMMCReadNWords(Uint32 *data, Uint32 numofBytes) 
{
  Uint16 i=0,j=0;
  Uint32 status;
  Uint32 fifoReadItrCount=0;          /* Word counter */
  fifoReadItrCount= numofBytes >> 5;

  for(i=0; i < fifoReadItrCount;i++)
  {			 	      
  	if(i != (fifoReadItrCount -1))
	{
	    
		// wait for DRRDY timeout
		do {
               status=SDMMC->MMCST0;
               if (status & 0xf8)
                   return (E_FAIL);
         } while(! (status & 0x0400) );	
	}
     
	for(j=0;j < 8; j++)
		*data++ = SDMMC->MMCDRR;
  }
  return ( E_PASS );
}

/**
    Send command to read a single block of data from the MMC/SD card

    cardAddr  Address of the location on the card to be read (should be a multiple of )
    dest  Buffer to read data into, from MMC/SD card
	blkLength  Size of the block in bytes, to be read from MMC/SD card

*/
Uint8 SDMMCSingleBlkRead( SDMMC_InfoHandle hSDMMCInfo, Uint32 cardMemAddr, Uint32 *dest, Uint32 blkLength ) 
{
  Uint8 status;
  Uint32 timeOut;

  /*Set the FIFO level 32 bytes (256 bit) always */
  SDMMC->MMCFIFOCTL |= 0x0004;
 
  status = SDMMCSendCmd(SDMMC_SET_BLOCKLEN, blkLength, FALSE);
  if(status != E_PASS)
  	return	E_FAIL;
      
  SDMMCClearResponse();

    /* Set the blocklength and size of the block in bytes */
  if(blkLength == 0)
    return E_FAIL;

  SDMMC->MMCBLEN |= (Uint16)(blkLength&0xFFF);
  SDMMC->MMCNBLK =  1;

    /* reset the FIFO  */
  SDMMC->MMCFIFOCTL |= 0x1;

  /* Set the Transfer direction from the FIFO as receive*/
  SDMMC->MMCFIFOCTL &= 0xFFFD;

  /*Send read single Block Command */       
  if(hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard)
    status = SDMMCSendCmd(0x2000 | SDMMC_READ_SINGLE_BLOCK, cardMemAddr >> hSDMMCInfo->dataBytesPerBlkPower2, FALSE);
  else
    status = SDMMCSendCmd(0x2000 | SDMMC_READ_SINGLE_BLOCK, cardMemAddr, FALSE);

  if(status != E_PASS)
	return E_FAIL;
		   		
  status=SDMMCReadNWords((Uint32*)dest, blkLength);
  if(status !=E_PASS)
	 return E_FAIL; 

  UTIL_waitLoop(100); 

  /*Delay needed for safety */
  timeOut = 3000;
  do {
	if(SDMMC->MMCST0 & SDMMC_STAT0_DATDNE) {
		return E_PASS;	
	}
	timeOut--;
	if( timeOut == 0 ) 
		return E_TIMEOUT;
  } while(1);

}
/**
    Send command to read multiple blocks of data from the MMC/SD card

    cardAddr  - Address of the location on the card to be read (should be a multiple of )
    dest      - Buffer to read data into, from MMC/SD card
	dataLength - Number of bytes to be read
*/
Uint8 SDMMCMultipleBlkRead( SDMMC_InfoHandle hSDMMCInfo, Uint32 cardMemAddr, Uint32 *dest, Uint32 dataLength) 
{
  Uint8 status, retVal;
  Uint32 numBlks;
  volatile Uint32 timeOut;

  retVal = E_PASS;

  /*Set the FIFO level 32 bytes (256 bit) always */
  SDMMC->MMCFIFOCTL |= 0x0004;
  status = SDMMCSendCmd(SDMMC_SET_BLOCKLEN, 512, FALSE);
  if(status != E_PASS)
  	return	E_FAIL;
     
  /* Clear all the response registers */
  SDMMCClearResponse();

  /* To find the Number of blocks to be read in terms of 512 Bytes */
  numBlks = dataLength >> hSDMMCInfo->dataBytesPerBlkPower2;

  SDMMC->MMCBLEN |= hSDMMCInfo->dataBytesPerBlk;
  SDMMC->MMCNBLK =  numBlks;

    /* reset the FIFO  */
  SDMMC->MMCFIFOCTL |= 0x1;

  /* Set the Transfer direction from the FIFO as receive*/
  SDMMC->MMCFIFOCTL &= 0xFFFD;

  /* Set the access width that will be used (MMCFIFOCTL register ACCWD bits) */

  /*Set the FIFO level(Threshold) 32 bytes (256 bit) always. After reset*/
  SDMMC->MMCFIFOCTL |= 0x0004;

  /* Send the Multi Blk Read Command with a no response check */
  if(hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard)
    status = SDMMCSendCmd(0xA080 | SDMMC_READ_MULTIPLE_BLOCK, cardMemAddr >> hSDMMCInfo->dataBytesPerBlkPower2, FALSE);
  else
    status = SDMMCSendCmd(0xA080 | SDMMC_READ_MULTIPLE_BLOCK, cardMemAddr, FALSE);

  if(status != E_PASS)
   	return E_FAIL;
	   		
  status=SDMMCReadNWords((Uint32*)dest, dataLength);
  if(status !=E_PASS)
	 return E_FAIL; 

  /*Delay needed for safety */
  UTIL_waitLoop(10000); 
  timeOut = 50000;
  do {
	if(SDMMC->MMCST0 & SDMMC_STAT0_DATDNE) {
	    retVal = E_PASS;
		break;
	}
	timeOut--;
	if( timeOut == 0 ) {
	    retVal = E_TIMEOUT;
		break;
	}
  } while(1);

  /* Send the Stop Tx Command */
  if(retVal == E_PASS) {
	  status = SDMMCSendCmd(SDMMC_STOP_TRANSMISSION, cardMemAddr, FALSE);
	  if(status != E_PASS)
	   	retVal = E_FAIL;
  }

  return retVal;
}

/**
    \brief  Write N words to the MMC Controller Register

    \param  data  Buffer containing data to be written to the MMC
    \param  numWords  Number of words to be written to the MMC

    \return if success, \c E_PASS, else error code

*/
Uint8 SDMMCWriteNWords( SDMMC_InfoHandle hSDMMCInfo, Uint32 *data, Uint32 numofBytes,Uint32 cardMemAddr ) 
{
  register volatile Uint16 stat;
  Uint32 ii=0,jj=0,fifoDxrWrCnt=0;
  Uint32 fifoWriteItrCount=0;          /* Word counter */
  Uint8 status = E_PASS;
#if 0
  /* reset the FIFO  */
  //CSL_FINS(SDMMC->MMCFIFOCTL,SDMMC_MMCFIFOCTL_FIFORST,1);
  SDMMC->MMCFIFOCTL = (((SDMMC->MMCFIFOCTL) & ~DEVICE_SDMMC_MMCFIFOCTL_FIFORST_MASK) | 
  ((1 << DEVICE_SDMMC_MMCFIFOCTL_FIFORST_SHIFT) & DEVICE_SDMMC_MMCFIFOCTL_FIFORST_MASK));
	
  /* Set the Transfer direction from the FIFO as transmit*/
  //CSL_FINS(SDMMC->MMCFIFOCTL,SDMMC_MMCFIFOCTL_FIFODIR,1);
  SDMMC->MMCFIFOCTL = (((SDMMC->MMCFIFOCTL) & ~DEVICE_SDMMC_MMCFIFOCTL_FIFODIR_MASK) | 
  ((1 << DEVICE_SDMMC_MMCFIFOCTL_FIFODIR_SHIFT) & DEVICE_SDMMC_MMCFIFOCTL_FIFODIR_MASK));
#endif
    /* reset the FIFO  */
  SDMMC->MMCFIFOCTL |= 0x1;

  /* Set the Transfer direction from the FIFO as receive*/
  SDMMC->MMCFIFOCTL |= 0x0002;

  /*Set the FIFO level 32 bytes (256 bit) always - Already done in Init TODO (remove?)*/
  SDMMC->MMCFIFOCTL |= 0x0004;

  fifoWriteItrCount=numofBytes >> 5;
  fifoDxrWrCnt=8;
     
  for(ii=0; ii < fifoWriteItrCount;ii++)
  {	
  	if(ii==0)
	{
		if(numofBytes == hSDMMCInfo->dataBytesPerBlk)
	        status = SDMMCSendCmd(0x2800 | SDMMC_WRITE_BLOCK, cardMemAddr,FALSE);
	    else
	        status = SDMMCSendCmd(0x2880 | SDMMC_WRITE_MULTIPLE_BLOCK, cardMemAddr, FALSE);
	}
  
  	for(jj=0;jj < fifoDxrWrCnt; jj++)
  		SDMMC->MMCDXR= (Uint32)*data++;

	   
  	if(ii != (fifoWriteItrCount -1))
  		while(!((SDMMC->MMCST0 & DEVICE_SDMMC_MMCST0_DXRDY_MASK) >> DEVICE_SDMMC_MMCST0_DXRDY_SHIFT));
  }	      		         		          		         		         
	     		        
  return (E_PASS);
}


Uint8 SDMMCSingleBlkWrite( SDMMC_InfoHandle hSDMMCInfo, Uint32 cardMemAddr, Uint32 *src, Uint32 blkLength)
{
  Uint8 status = E_PASS;

  SDMMC->MMCBLEN |= (Uint16)(hSDMMCInfo->dataBytesPerBlk&0xFFF);
  SDMMC->MMCNBLK =  1;
    
  status = SDMMCSendCmd(SDMMC_SET_BLOCKLEN, hSDMMCInfo->dataBytesPerBlk, FALSE);
  if(status != E_PASS)
  	return	E_FAIL;             
 
  if(status == E_PASS) 
  {
    if(hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard)
	{
  	  /* Write Data, every time MMCDXR Reg full */
	  status = SDMMCWriteNWords(hSDMMCInfo,(Uint32*)src, hSDMMCInfo->dataBytesPerBlk,cardMemAddr >> hSDMMCInfo->dataBytesPerBlkPower2);
	}
	else
	{
  	  /* Write Data, every time MMCDXR Reg full */
	  status = SDMMCWriteNWords(hSDMMCInfo, (Uint32*)src, hSDMMCInfo->dataBytesPerBlk,cardMemAddr);
	}
		   
	/*Delay Required*/
	//UTIL_waitLoop(100);
		 
    if(status == E_PASS)
    {
	do {
	  	if(SDMMC->MMCST0 & SDMMC_STAT0_DATDNE)
			return E_PASS;
	} while(1);       
		   	     
    }else
		 return E_FAIL;
  }	
  return E_PASS;
}

Uint8 SDMMCMultipleBlkWrite( SDMMC_InfoHandle hSDMMCInfo, Uint32 cardMemAddr, Uint32 *src, Uint32 datalength) 
{
	Uint8 status;
	Uint32 numBlks;
      
   /*Clear all Response Registers */  
   SDMMCClearResponse();
     
   /* To find out number of BLOCKS to be written in terms of 512 bytes */
   numBlks = datalength >> hSDMMCInfo->dataBytesPerBlkPower2;

   if(numBlks* hSDMMCInfo->dataBytesPerBlk < datalength)
     numBlks++;
    
   SDMMC->MMCBLEN |= hSDMMCInfo->dataBytesPerBlk;
   SDMMC->MMCNBLK =  numBlks;

   /* Send the BLOCK LEN command to the card , it is an optional one*/
   status = SDMMCSendCmd(SDMMC_SET_BLOCKLEN, hSDMMCInfo->dataBytesPerBlk, FALSE);
   if(status!=E_PASS)
	   return E_FAIL;  
    if(hSDMMCInfo->mmcSdCSDRegInfo.sdHighCapacityCard)
	{             		
      /* Read Data, every time Data Rd Reg full, till all words read */
      status = SDMMCWriteNWords(hSDMMCInfo, (Uint32*)src,datalength,cardMemAddr >> hSDMMCInfo->dataBytesPerBlkPower2);
	}
    else
	{             		
      /* Read Data, every time Data Rd Reg full, till all words read */
      status = SDMMCWriteNWords(hSDMMCInfo, (Uint32*)src,datalength,cardMemAddr);
	}

   /*Delay required */
   UTIL_waitLoop(100);
			  
   if(status != E_PASS)
	return E_FAIL;

   /* Check for SDMMC_DATDNE signal */ 
   /* No timeout in QT mode */
   do {
   	if(SDMMC->MMCST0 & SDMMC_STAT0_DATDNE) {
 		SDMMCSendCmd(0xA180 | SDMMC_STOP_TRANSMISSION, SDMMC_STUFF_BITS, FALSE);
		    return E_PASS;
	}
   } while(1);       
}


/***********************************************************
* End file                                                 *
***********************************************************/

