/*
 * FormUartHost.cs
*/

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
*/

/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/
// FormUartHost.cs ------------------------------------------------------------------------------
//
// Implements GUI wrapper to run UART boot using AIS_Parser class (in AIS_Util namespace).
//
// Written by Joe Coombs, 2009-02-20
//
// Copyright (C) 2008 Texas Instruments Incorporated
// All Rights Reserved
// ------------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;
using System.Threading;
using TI.AISLib;

namespace TIBootAndFlash
{
  delegate void CloseDelegate();

  public partial class FormUartHost : Form
  {
    private SerialPort myUART;
    private Thread bootThread;
    private int baud;
    private byte[] ais;

    public FormUartHost()
    {
      InitializeComponent();
      comboBoxBaud.SelectedIndex = 0;
    }

    public void UTIL_log(String text)
    {
      // print text as new line in log text box (append newline)
      richTextBoxLog.AppendText(text + "\n");
      // scroll box to end
      richTextBoxLog.ScrollToCaret();
      // force redraw
      richTextBoxLog.Refresh();
    }

    // invoke version; can be called safely from another thread
    public void UTIL_log_INVOKE(String text)
    {
      LogFxnDelegate temp_delegate = new LogFxnDelegate(UTIL_log);
      this.Invoke(temp_delegate, new object[] { text });
    }

    public void UTIL_clr_log()
    {
      // reset log box contents
      richTextBoxLog.Clear();
    }

    private AisStatus UTIL_uartRead(byte[] rcvBuf, int index, int rcvSize, int timeout)
    {
      int bytesRead = 0;

      myUART.ReadTimeout = timeout;
      try
      {
        // note: this is necessary because the SerialPort.Read method likes to return
        // after reading only one byte, regardless of the specified count
        while (bytesRead < rcvSize)
        {
          bytesRead += myUART.Read(rcvBuf, index + bytesRead, rcvSize - bytesRead);
        }
      }
      catch (Exception ex)
      {
        // log using invoke since this will be called by other thread
        UTIL_log_INVOKE("(Serial Port): Read error! (" + ex.Message + ")");
        return AisStatus.ERROR;
      }

      return AisStatus.IN_PROGRESS;
    }

    private AisStatus UTIL_uartWrite(byte[] xmtBuf, int index, int xmtSize, int timeout)
    {
      myUART.WriteTimeout = timeout;
      try
      {
        myUART.Write(xmtBuf, index, xmtSize);
      }
      catch (Exception ex)
      {
        // log using invoke since this will be called by other thread
        UTIL_log_INVOKE("(Serial Port): Write error! (" + ex.Message + ")");
        return AisStatus.ERROR;
      }

      return AisStatus.IN_PROGRESS;
    }

    private void UTIL_boot()
    {
      // create and run AIS parser
      AIS_Parser parser = new AIS_Parser(AisHostType.UART, AisSecureType.NONE, this.UTIL_log_INVOKE, this.UTIL_uartRead, this.UTIL_uartWrite);
      parser.waitBOOTME = checkBoxWaitForBootme.Checked;

      try
      {
        parser.boot(ais);

        // reset form when done (use invoke)
        UTIL_close_boot_INVOKE();
      }
      catch (Exception ex)
      {
        UTIL_log_INVOKE("(System): Unhandled exception (" + ex.Message + ")");
      }
    }

    private void UTIL_close_boot()
    {
      // close serial port
      try
      {
        myUART.Close();

        UTIL_log("(Serial Port): Closing " + textBoxPort.Text + ".");

        // toggle on buttons
        buttonStart.Enabled = true;
        buttonStop.Enabled = false;
      }
      catch (Exception ex)
      {
        UTIL_log("(System): Unhandled exception (" + ex.Message + ")");
      }
    }

    // invoke version; can be called safely from another thread
    private void UTIL_close_boot_INVOKE()
    {
      CloseDelegate temp_delegate = new CloseDelegate(UTIL_close_boot);
      this.Invoke(temp_delegate, new object[] { });
    }

    private void buttonStart_Click(object sender, EventArgs e)
    {
      // clear log
      UTIL_clr_log();

      // dump contents of specified AIS file
      try
      {
        ais = File.ReadAllBytes(textBoxFileName.Text);
        UTIL_log(String.Format("(File IO): Read {0} bytes from file ", ais.Length) + textBoxFileName.Text + ".");
      }
      catch (Exception ex)
      {
        UTIL_log("(File IO): Error reading file \"" + textBoxFileName.Text + "\". (" + ex.Message + ")");
        return;
      }
        
      // open serial port
      try
      {
         UTIL_log("(Serial Port): Opening " + textBoxPort.Text + " at " + baud.ToString() + " baud...");
         myUART = new SerialPort(textBoxPort.Text, baud, Parity.None, 8, StopBits.One); 
         myUART.Open();
      }
      catch (Exception ex)
      {
         UTIL_log("(Serial Port): Error opening " + textBoxPort.Text + "! (" + ex.Message + ")");
         return;
      }

      // toggle buttons before processing
      buttonStart.Enabled = false;
      buttonStop.Enabled = true;

      // parse AIS in new thread
      bootThread = new Thread(UTIL_boot);
      bootThread.Start();
    }

    private void buttonStop_Click(object sender, EventArgs e)
    {
      // make sure boot is running
      if (bootThread.IsAlive == false)
          return;

      // abort boot
      bootThread.Suspend();
      // Note: Thread.Abort does not play nicely with SerialPort operations.
      //       Thread.Suspend is deprecated and messy (we never dispose the thread),
      //       but it'll have to do.
      UTIL_log("(System): Boot aborted by user.");
      UTIL_close_boot();
    }

    private void buttonFileBrowse_Click(object sender, EventArgs e)
    {
      // open file browser
      OpenFileDialog aisOpenSel = new OpenFileDialog();

      aisOpenSel.Filter = "AIS binary files|*.bin;*.ais|All files|*.*";
      aisOpenSel.Title = "Open AIS File";
      try
      {
          if (aisOpenSel.ShowDialog() == DialogResult.OK)
              textBoxFileName.Text = aisOpenSel.FileName;
      }
      catch (Exception ex)
      {
          UTIL_log("(System): Error retrieving file name (" + ex.Message + ")");
          textBoxFileName.Clear();
      }
    }

    private void comboBoxBaud_SelectedIndexChanged(object sender, EventArgs e)
    {
        try
        {
            baud = Convert.ToInt32(comboBoxBaud.Text);

            // reset color
            comboBoxBaud.ForeColor = Color.Black;
        }
        catch
        {
            // set baud to default and warn user
            baud = 115200;
            comboBoxBaud.ForeColor = Color.Red;
        }
    }

    private void FormUartHost_Load(object sender, EventArgs e)
    {

    }
  }
}